﻿using System;
using System.Collections.Generic;
using System.ComponentModel;
using System.Diagnostics;
using System.Linq;
using System.Threading.Tasks;
using Buddy.Coroutines;
using Clio.Utilities;
using ff14bot;
using ff14bot.AClasses;
using ff14bot.Behavior;
using ff14bot.Enums;
using ff14bot.Helpers;
using ff14bot.Managers;
using ff14bot.Navigation;
using ff14bot.Objects;
using ff14bot.RemoteWindows;
using ff14bot.Settings;
using TreeSharp;
using Action = TreeSharp.Action;

namespace RawMeat
{
    public class RawMeatRunner : BotBase
    {
        private static FateMapper _mapper;
        private static FateData _focusedFate;
        private static readonly Stopwatch LastFateIdTimer = new Stopwatch();
        private static uint _lastFateId;
        private static readonly Random Random = new Random();
        private static Composite _root;
        private static IEnumerable<BattleCharacter> _currentFateMobs;
        private static IOrderedEnumerable<FateData> _activeFates;

        private static readonly Dictionary<FateIconType, float> FateScore = new Dictionary<FateIconType, float>
        {
            {FateIconType.Boss, 0.5f},
            {FateIconType.Battle, 1.0f},
            {FateIconType.KillHandIn, 1.0f},
            {FateIconType.ProtectNPC, 1.5f},
            {FateIconType.ProtectNPC2, 1.5f}
        };

        private static readonly Stopwatch ClusterTimer = Stopwatch.StartNew();

        private static uint LastFateId
        {
            get
            {
                if (LastFateIdTimer.ElapsedMilliseconds > 1000*45)
                    _lastFateId = 0;
                return _lastFateId;
            }
            set
            {
                LastFateIdTimer.Restart();
                _lastFateId = value;
            }
        }

        [Localizable(false)]
        public override string Name
        {
            get { return "RawMeat"; }
        }

        public override PulseFlags PulseFlags
        {
            get { return PulseFlags.All; }
        }

        public override bool IsAutonomous
        {
            get { return true; }
        }

        public override bool WantButton
        {
            get { return true; }
        }

        public override Composite Root
        {
            get
            {
                return _root ?? (_root =
                    new Sequence(
                        new PrioritySelector(
                            // Set Poi on death
                            new Decorator(r => Core.Me.IsDead,
                                new Action(r => Poi.Current = new Poi(Core.Me, PoiType.Death))),
                            // Check if something that we are attacking is a fate monster or not.
                            //new Decorator(r => Poi.Current.Type == PoiType.Kill && Poi.Current.Unit.IsValid && Poi.Current.BattleCharacter.IsFate == false && Poi.Current.BattleCharacter.IsFateGone == false,
                            //        new Action(r => Poi.Clear("Not attacking that!"))
                             //       ),
                            // Check if the fate is still valid or not, if not then set lastFateId to it and remove it.
                            new Decorator(r =>
                            {
                                var orDefault =
                                    _activeFates.FirstOrDefault();
                                return _focusedFate != null &&
                                       (!_focusedFate.IsValid ||
                                        (orDefault != null && _focusedFate.Id != orDefault.Id &&
                                         _mapper.FateChainMaps.All(chain => _focusedFate.Id != chain[0])));
                            }, new Action(r =>
                            {
                                Logger.Log("Fate {0} ended", _focusedFate.Name);
                                LastFateId = _focusedFate.Id;
                                _focusedFate = null;
                                return RunStatus.Success;
                            }
                                )),
                            // Get new fate or go back home
                            new Decorator(r => _focusedFate == null,
                                new PrioritySelector(
                                    // Get new fate
                                    new Decorator(r => _activeFates.Any(),
                                        new Action(
                                            r =>
                                            {
                                                _focusedFate = _activeFates.FirstOrDefault();

                                                if (WaitForNextFate() && !IsNextFate(_focusedFate))
                                                {
                                                    Logger.Log("Waiting for the next fate in chain");
                                                    _focusedFate = null;
                                                    Navigator.MoveToPointWithin(
                                                        _mapper.FateLocationMaps.First(f => f.FateId == LastFateId)
                                                            .FateLocation, 10);
                                                }
                                                else
                                                {
                                                    if (_focusedFate != null)
                                                        Logger.Log("Found a new fate: {0}", _focusedFate.Name);
                                                }

                                                return RunStatus.Success;
                                            }
                                            )
                                        ),
                                    // Go home, check if we are not waiting for next fate.
                                    new Decorator(
                                        r =>
                                            !_activeFates.Any() && Poi.Current.Type != PoiType.Kill &&
                                            !WaitForNextFate(),
                                        new ActionRunCoroutine(async r =>
                                        {
                                            Logger.Log("No fate found. Moving to the nearest Aetheryte");
                                            var destination = WorldManager.AetheryteIdsForZone(WorldManager.ZoneId)
                                                .Select(a => a.Item2)
                                                .OrderBy(a => Core.Me.Distance(a))
                                                .FirstOrDefault();

                                            if (Core.Me.Distance(destination) > CharacterSettings.Instance.MountDistance &&
                                                !Core.Me.IsMounted && Actionmanager.CanMount == 0 &&
                                                CharacterSettings.Instance.UseMount)
                                            {
                                                await Mount();
                                            }

                                            // The destination doesn't change so we can call this. Also calling MoveAndStop will act weird because it can't detect objects (like walking into the aetheryte).
                                            Navigator.MoveToPointWithin(destination, 30);

                                            return true;
                                        })
                                        )
                                    )
                                ),

                            // We have a fate to go to
                            new Decorator(r => _focusedFate != null && _focusedFate.IsValid,
                                new PrioritySelector(
                                    // If within a fate near you and level is not sync.
                                    // We need nearby fate incase it overlaps, it will toggle repeatly if level sync max fate level != current fate max fate level
                                    new Decorator(
                                        r =>
                                        {
                                            var fateData =
                                                FateManager.ActiveFates.Where(
                                                    fate => Core.Me.Distance(fate.Location) < fate.Radius)
                                                    .OrderBy(fate => fate.MaxLevel)
                                                    .FirstOrDefault();
                                            return fateData != null && (FateManager.WithinFate &&
                                                                        fateData
                                                                            .MaxLevel < Core.Player.ClassLevel);
                                        },
                                        new ActionRunCoroutine(async r =>
                                        {
                                            Logger.Log("Applying Level Sync");

                                            ToDoList.LevelSync();

                                            await Coroutine.Sleep(500);

                                            return true;
                                        })
                                        ),

                                    // 
                                    // Are we in the targeted fate?
                                    new Decorator(r =>
                                    {
                                        var firstOrDefault =
                                            _activeFates
                                                .FirstOrDefault();
                                        return firstOrDefault != null &&
                                               (FateManager.WithinFate && firstOrDefault.Id == _focusedFate.Id &&
                                                _focusedFate.MaxLevel >= Core.Me.ClassLevel);
                                    },
                                        new PrioritySelector(
                                            // Find a fate monster to kill
                                            new Decorator(
                                                r => Poi.Current.Type != PoiType.Kill && _currentFateMobs.Any(),
                                                new Action(
                                                    r =>
                                                    {
                                                        var closesMob =
                                                            _currentFateMobs.OrderByDescending(mob => mob.MaxHealth)
                                                                .ThenBy(mob => Core.Me.Distance(mob.Location))
                                                                .FirstOrDefault();
                                                        Logger.Log("Targetting a fate mob {0}", closesMob);
                                                        // Order by HP, we want to attack the boss first.
                                                        Poi.Current =
                                                            new Poi(closesMob
                                                                ,
                                                                PoiType.Kill);
                                                    }
                                                    )
                                                )
                                            )
                                        ),
                                    new Decorator(r => Poi.Current.Type != PoiType.Kill,
                                        new ActionRunCoroutine(async r =>
                                        {
                                            if (_focusedFate != null)
                                                if (
                                                    _focusedFate.Icon ==
                                                    FateIconType.ProtectNPC ||
                                                    _focusedFate.Icon == FateIconType.ProtectNPC2)
                                                {
                                                    var npc = GameObjectManager
                                                        .GetObjectsOfType<BattleCharacter>()
                                                        .FirstOrDefault(
                                                            b => b.IsFate && !b.CanAttack && b.FateId == _focusedFate.Id);

                                                    if (npc != null && npc.IsValid && (npc.IsBehind || npc.IsFlanking) &&
                                                        Core.Me.Distance(npc) > 7)
                                                    {
                                                        Logger.Log("Moving using escort fate logic");
                                                        Navigator.Stop();

                                                        Func<bool> isInFront = () => !npc.IsBehind;

                                                        await Coroutine.Sleep(Random.Next(0, 1000));
                                                        MovementManager.MoveForwardStart();
                                                        while (npc.IsValid && (Core.Me.Distance(npc) > 1 || !isInFront()))
                                                        {
                                                            Core.Me.Face(npc);
                                                            if (!MovementManager.IsMoving)
                                                                MovementManager.MoveForwardStart();
                                                            await Coroutine.Sleep(200);
                                                        }

                                                        await Coroutine.Sleep(Random.Next(500, 1000));
                                                        MovementManager.MoveForwardStop();

                                                        Logger.Log("Reached destination, moving stopped");
                                                    }
                                                    else
                                                    {
                                                        if (FateManager.WithinFate)
                                                        {
                                                            Logger.Log("Idle in escort fate.");
                                                        }
                                                        else
                                                        {
                                                            await MoveToFocusedFate();
                                                        }
                                                    }
                                                }
                                                else
                                                {
                                                    await MoveToFocusedFate();
                                                }
                                            return true;
                                        })
                                        )
                                    )
                                ),
                            new ActionAlwaysSucceed()
                            ),
                        new Decorator(r => Poi.Current.Type != PoiType.None, BrainBehavior.CreateBrain())
                        )
                    );
            }
        }

        public override bool RequiresProfile
        {
            get { return false; }
        }

        private static async Task MoveToFocusedFate()
        {
            Vector3 currentMove;
            if (FateManager.WithinFate &&
                FateManager.ActiveFates.Where(fate => Core.Me.Distance(fate.Location) < fate.Radius)
                    .Any(fate => fate.Id == _focusedFate.Id))
            {
                if (ClusterTimer.ElapsedMilliseconds > 5000)
                {
                    Logger.Log("Moving using cluster logic");

                    var x = 0.0f;
                    var y = 0.0f;
                    var z = 0.0f;
                    var total = 0.0f;
                    GameObjectManager.GetObjectsOfType<BattleCharacter>()
                        .Where(
                            bc =>
                                bc.Type == GameObjectType.Pc &&
                                bc.Location.Distance(_focusedFate.Location) < _focusedFate.Radius)
                        .ForEach(
                            bc =>
                            {
                                total++;
                                x += bc.Location.X;
                                y += bc.Location.Y;
                                z += bc.Location.Z;
                            });
                    currentMove = new Vector3(x/total, y/total, z/total);
                    Navigator.MoveTo(currentMove);

                    ClusterTimer.Restart();
                }
                else
                {
                    Logger.Log("Using cluster cache");
                }
            }
            else
            {
                Logger.Log("Moving using centre logic");

                currentMove = MapFateLocation(_focusedFate);

                if (Core.Me.Distance(currentMove) >
                    CharacterSettings.Instance.MountDistance && !Core.Me.IsMounted &&
                    Actionmanager.CanMount == 0 && CharacterSettings.Instance.UseMount)
                {
                    Logger.Log("Distance too far, using mount");
                    await Mount();
                }

                Navigator.MoveToPointWithin(currentMove, 15);
            }
        }

        public override void Pulse()
        {
            base.Pulse();

            // Fate mobs that is in the current fate.
            _currentFateMobs = GameObjectManager.GetObjectsOfType<BattleCharacter>()
                .Where(
                    b =>
                        b.IsFate && !b.IsFateGone && b.CanAttack && b.FateId == _focusedFate.Id);

            // Order by Boss > Others > Protect/Escort > Distance
            //_activeFates = FateManager.ActiveFates.OrderByDescending(f => f.Icon == FateIconType.Boss).ThenBy(f => f.Icon == FateIconType.ProtectNPC || f.Icon == FateIconType.ProtectNPC2).ThenBy(f => Core.Me.Distance(f.Location));

            _activeFates =
                FateManager.ActiveFates.OrderBy(fate => _mapper.FateChainMaps.All(chain => fate.Id != chain[0]))
                    .ThenByDescending(
                        fate =>
                        {
                            float fateScore;
                            fateScore = FateScore.TryGetValue(fate.Icon, out fateScore) ? fateScore : 1.0f;
                            return (30000 -
                                    fateScore*
                                    (Math.Pow(fate.Progress, 2) + Core.Me.Distance(fate.Location)*100));
                        });
        }

        private static async Task<bool> Mount()
        {
            var ticks = 0;
            while (ticks < 20 && !Core.Me.IsMounted && !Core.Me.InCombat)
            {
                if (MovementManager.IsMoving) Navigator.Stop();
                if (!Core.Me.IsCasting) Actionmanager.Mount();
                await Coroutine.Sleep(200);
                ticks++;
            }

            return Core.Me.IsMounted;
        }

        public override void OnButtonPress()
        {
        }

        private static bool WaitForNextFate()
        {
            return _mapper.FateChainMaps.Any(chain => chain[0] == LastFateId);
        }

        private static bool IsNextFate(FateData fateData)
        {
            return _mapper.FateChainMaps.Any(id => id[0] == LastFateId && id[1] == fateData.Id);
        }

        private static Vector3 MapFateLocation(FateData fateData)
        {
            var currentDestination = GetFateLocation(fateData);
            return currentDestination;
        }

        private static Vector3 GetFateLocation(FateData fateData)
        {
            var firstOrDefault = _mapper.FateLocationMaps
                .FirstOrDefault(id => id.FateId == fateData.Id);
            if (firstOrDefault != null)
            {
                return
                    firstOrDefault
                        .FateLocation;
            }

            return fateData.Location;
        }

        public override void Initialize()
        {
            Logger.Log("Initializing");
        }

        public override void Start()
        {
            Logger.Log("Starting");

            Navigator.PlayerMover = new SlideMover();
            if (!(Navigator.NavigationProvider is GaiaNavigator))
            {
                Navigator.NavigationProvider = new GaiaNavigator();
            }
            else
            {
                Navigator.NavigationProvider.PathPrecision = 0.8f;
            }

            GameSettingsManager.FaceTargetOnAction = true;
            GameSettingsManager.FlightMode = true;
            
            _mapper = new FateMapper
            {
                FateChainMaps = new[]
                {
                    // Dark Devices
                    new uint[] {643, 644},
                    new uint[] {644, 645},
                    new uint[] {645, 646},

                    // Svara 
                    new uint[] {501, 502},
                    new uint[] {502, 503},
                    new uint[] {502, 504}
                },
                FateLocationMaps = new[]
                {
                    new FateMapper.LocationMap(643, new Vector3("252.831238, 25, 4.66978")),
                    new FateMapper.LocationMap(644, new Vector3("252.831238, 25, 4.66978")),
                    new FateMapper.LocationMap(645, new Vector3("252.831238, 25, 4.66978")),
                    new FateMapper.LocationMap(646, new Vector3("252.831238, 25, 4.66978")),
                    new FateMapper.LocationMap(637, new Vector3("183.645386, 23.9917164, 107.8395"))
                },
                Version = 4
            };
            //Logging.Write(Encoding.Base64Encode(JsonConvert.SerializeObject(t)));

            // Pulse once before the tick happens so we have all data.
            Pulse();
        }

        public override void Stop()
        {
            base.Stop();
            //Navigator.NavigationProvider = null;
            TreeHooks.Instance.ClearAll();
        }
    }
}